--[[
Name: Quixote-1.0
Revision: $Revision: 23082 $
Author(s): Kemayo (kemayo@gmail.com)
Website: http://www.wowace.com/wiki/Quixote-1.0
Documentation: http://www.wowace.com/wiki/Quixote-1.0
SVN: http://svn.wowace.com/wowace/trunk/Quixote/
Description: Abstracts out questlog handling.
Dependencies: AceLibrary, AceEvent, AceComm, Deformat
]]

local MAJOR_VERSION = "Quixote-1.0"
local MINOR_VERSION = "$Revision: 23082 $"

if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

if not AceLibrary:HasInstance("AceEvent-2.0") then error(MAJOR_VERSION .. " requires AceEvent-2.0.") end
if not AceLibrary:HasInstance("AceComm-2.0") then error(MAJOR_VERSION .. " requires AceComm-2.0.") end
if not AceLibrary:HasInstance("Deformat-2.0") then error(MAJOR_VERSION .. " requires Deformat-2.0.") end

local deformat
local Quixote = {}

local oldandbusted = _G.GetBuildInfo() == '2.0.1'

------------------------------------------------
-- Light localization
------------------------------------------------

local tag = {}
local DUNGEON = LFG_TYPE_DUNGEON
local HEROIC = DUNGEON_DIFFICULTY2
local PVP = PVP_ENABLED

tag[ELITE] = '+'
tag[GROUP] = '+'
tag[PVP] = 'p'
tag[RAID] = 'r'
tag[DUNGEON] = 'd'
tag[HEROIC] = 'd+'

------------------------------------------------
-- activate, external
------------------------------------------------

local function activate(self, oldLib, oldDeactivate)
	if oldLib then
		self.quests = oldLib.quests
		self.quests_zone = oldLib.quests_zone
		self.quests_level = oldLib.quests_level
		self.mobs = oldLib.mobs
		self.items = oldLib.items
		self.quest_names = oldLib.quest_names
		self.zones = oldLib.zones
		self.numQuests = oldLib.numQuests
		self.numQuestsDone = oldLib.numQuestsDone
		self.firstScanDone = oldLib.firstScanDone
		self.party = oldLib.party
		
		oldLib:UnregisterAllEvents()
		oldLib:CancelAllScheduledEvents()
		
		if oldLib.UnregisterComm then
			oldLib:UnregisterAllComms()
		end
	end
	
	if not self.quests then self.quests = {} end
	if not self.quests_zone then self.quests_zone = {} end
	if not self.quests_level then self.quests_level = {} end
	if not self.mobs then self.mobs = {} end
	if not self.items then self.items = {} end
	if not self.quest_names then self.quest_names = {} end
	if not self.zones then self.zones = {} end
	if not self.numQuests then self.numQuests = 0 end
	if not self.numQuestsDone then self.numQuestsDone = 0 end
	if not self.firstScanDone then self.firstScanDone = false end
	if not self.party then self.party = {} end
	
	if oldDeactivate then oldDeactivate(oldLib) end
end

local function external(self, major, instance)
	if major == "AceEvent-2.0" then
		instance:embed(self)
		self:UnregisterAllEvents()
		self:CancelAllScheduledEvents()
		if instance:IsFullyInitialized() then
			self:AceEvent_FullyInitialized()
		else
			self:RegisterEvent("AceEvent_FullyInitialized", "AceEvent_FullyInitialized", true)
		end
	elseif major == "Deformat-2.0" then
		deformat = instance
	elseif major == "AceComm-2.0" then
		instance:embed(self)
		self:SetCommPrefix("Quixote")
		self:RegisterComm(self.commPrefix, "PARTY")
		self:RegisterComm(self.commPrefix, "WHISPER")
	end
end

------------------------------------------------
-- Internal functions
------------------------------------------------

function Quixote:AceEvent_FullyInitialized()
	self:RegisterBucketEvent("QUEST_LOG_UPDATE", 1)
	--We use a bucket event because scanning the quest log can trigger QUEST_LOG_UPDATE once per quest-related item if it has to repopulate the game's itemcache.
	self:QUEST_LOG_UPDATE()
	
	self:TriggerEvent("Quixote_Enabled")
	--Comm stuff:
	self:RegisterEvent("Quixote_Leaderboard_Update")
	self:RegisterEvent("Quixote_Quest_Gained")
	self:RegisterEvent("Quixote_Quest_Lost")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:PARTY_MEMBERS_CHANGED()
end

local function tableCleaner(t)
	for k in pairs(t) do
		if type(t[k]) == 'table' then
			tableCleaner(t[k])
		end
		t[k] = nil
	end
end

------------------------------------------------
-- Quest scanning
------------------------------------------------

function Quixote:QUEST_LOG_UPDATE()
	local startingQuestLogSelection = GetQuestLogSelection()
	local numEntries, numQuests = GetNumQuestLogEntries()
	local numQuestsDone = 0
	
	local quests = {}
	local quests_level = {}
	local quests_zone = {}
	local items = {}
	local mobs = {}
	local zones = {}
	local quest_names = {}
	local zoneIndex = "Unknown" --The first item in the quest log should always be a header, so this should always be replaced.  But just in case, give it a default value.
	
	if numEntries > 0 then
		local questid
		for questid = 1, numEntries do
			SelectQuestLogEntry(questid) -- this is required for a few of the quest functions.
			local strQuestLogTitleText, strQuestLevel, strQuestTag, isHeader, isCollapsed, isComplete
			if oldandbusted then
				-- This can be removed when the 2.0.2 patch hits live.
				strQuestLogTitleText, strQuestLevel, strQuestTag, isHeader, isCollapsed, isComplete = GetQuestLogTitle(questid)
			else
				strQuestLogTitleText, strQuestLevel, strQuestTag, suggestedGroup, isHeader, isCollapsed, isComplete = GetQuestLogTitle(questid)
			end
			
			if isHeader then
				zoneIndex = strQuestLogTitleText
				if not quests_zone[zoneIndex] then
					table.insert(zones, zoneIndex)
					quests_zone[zoneIndex] = {}
				end
			else
				local q = {}
				local strQuestDescription, strQuestObjectives = GetQuestLogQuestText(questid)
				
				if string.find(strQuestLogTitleText, '^%[') then _,_,strQuestLogTitleText = string.find(strQuestLogTitleText, "^%[?%d*.?]?%s?(.*)") end
				
				quest_names[strQuestLogTitleText] = questid
				
				q.zone = zoneIndex
				q.id = questid
				q.title = strQuestLogTitleText
				q.level = strQuestLevel
				q.tag = tag[strQuestTag] and tag[strQuestTag] or ''
				q.complete = isComplete
				--q.description = strQuestDescription
				q.objective = strQuestObjectives
				
				if isComplete == 1 then
					numQuestsDone = numQuestsDone + 1
				end
				
				local lb = {}
				if GetNumQuestLeaderBoards(questid) > 0 then
					local ii
					for ii=1, GetNumQuestLeaderBoards(questid) do
						local desc, qtype, done = GetQuestLogLeaderBoard(ii, questid)
						local numNeeded,numItems
						local l = {}
						
						if qtype == 'item' or qtype == 'object' then --'object' for the leaderboard in Dousing the Flames of Protection; maybe others.
							desc,numItems,numNeeded = deformat(desc, QUEST_OBJECTS_FOUND)
							items[desc] = string.format("%s: %d/%d", strQuestLogTitleText, numItems, numNeeded)
						elseif qtype == 'monster' then
							mobName,numItems,numNeeded = deformat(desc, QUEST_MONSTERS_KILLED)
							if mobName == nil or numItems == nil or numNeeded == nil then
								--Sometimes we get objectives like "Find Mankrik's Wife: 0/1", which are listed as "monster".
								mobName, numItems, numNeeded = deformat(desc, QUEST_OBJECTS_FOUND)
							end
							desc = mobName
							mobs[desc] = string.format("%s: %d/%d", strQuestLogTitleText, numItems, numNeeded)
						elseif qtype == 'reputation' then
							desc,numItems,numNeeded = deformat(desc, QUEST_FACTION_NEEDED)
						elseif qtype == 'event' then
							numNeeded = 1
							if done then numItems = 1 else numItems = 0 end
						end
						
						l.description = desc
						l.type = qtype
						l.numGot = numItems
						l.numNeeded = numNeeded
						l.done = done
						
						table.insert(lb, l)
					end
				end
				q.leaderboard = lb
				
				quests[questid] = q
				table.insert(quests_zone[zoneIndex], questid)
				table.insert(quests_level, questid)
			end
		end
		SelectQuestLogEntry(startingQuestLogSelection)
	end
	
	local oldquests, oldquest_names = self.quests, self.quest_names
	tableCleaner(self.quests_zone); tableCleaner(self.quests_level); tableCleaner(self.mobs); tableCleaner(self.items); tableCleaner(self.zones)
	
	self.quests = quests
	self.quests_level = self:SortQuests(quests_level)
	for zone, quests in pairs(quests_zone) do
		quests_zone[zone] = self:SortQuests(quests)
	end
	self.quests_zone = quests_zone
	self.items = items
	self.mobs = mobs
	self.zones = zones
	self.numQuests = numQuests
	self.numEntries = numEntries
	self.numQuestsDone = numQuestsDone
	self.quest_names = quest_names
	
	if self.firstScanDone then
		--Do all this checking down here so we can guarantee that library calls in response to triggered events will receive data about the current state of the quest log.
		--have we gained any quests?
		for name,id in pairs(quest_names) do
			if not oldquest_names[name] then
				self:TriggerEvent("Quixote_Quest_Gained", name, id, table.getn(quests[id].leaderboard))
			else
				for l, goal in pairs(quests[id].leaderboard) do
					local oldid = oldquest_names[name]
					if oldquests[oldid].leaderboard[l].description == goal.description and oldquests[oldid].leaderboard[l].numGot ~= goal.numGot then
						self:TriggerEvent("Quixote_Leaderboard_Update", name, id, l, goal.description, oldquests[oldid].leaderboard[l].numGot, goal.numGot, goal.numNeeded, goal.type)
					end
				end
			end
		end
		--have we lost any quests?
		for name, id in pairs(oldquest_names) do
			if not quest_names[name] then
				self:TriggerEvent("Quixote_Quest_Lost", name, id)
			end
		end
	end
	
	tableCleaner(oldquests)
	tableCleaner(oldquest_names)
	
	if not self.firstScanDone then
		self.firstScanDone = true
	end
	
	self:TriggerEvent("Quixote_Update")
end

-- Sorts a table of quests by level, with quests of the same level ordered
-- by elite, dungeon or raid tags, i.e. normal < elite < dungeon < raid.
-- Quests of the same level and tag are sorted alphabetically by title.
-- Returns a new table of the sorted quests. Original table is unchanged.
local mySort
function Quixote:SortQuests(quests)
	if not mySort then
		mySort = function(a,b)
			local q = self.quests
			local aa = q[a].level*4
			local bb = q[b].level*4
			if q[a].tag == tag[ELITE] then aa = aa+1
			elseif q[a].tag == tag[GROUP] then aa = aa+1
			elseif q[a].tag == tag[DUNGEON] then aa = aa+2
			elseif q[a].tag == tag[HEROIC] then aa = aa+3
			elseif q[a].tag == tag[RAID] then aa = aa+4
			elseif q[a].tag == tag[PVP] then aa = aa+5 end
			if q[b].tag == tag[ELITE] then bb = bb+1
			elseif q[b].tag == tag[GROUP] then bb = bb+1
			elseif q[b].tag == tag[DUNGEON] then bb = bb+2
			elseif q[b].tag == tag[HEROIC] then bb = bb+3
			elseif q[b].tag == tag[RAID] then bb = bb+4
			elseif q[b].tag == tag[PVP] then bb = bb+5 end
			if aa == bb then
				return q[a].title < q[b].title
			end
			return aa < bb;
		end
	end
	table.sort(quests, mySort)

	return quests
end

local function questIter(questList, i)
	i = i + 1
	v = questList[i]
	if v then
		return i, v
	end
end

------------------------------------------------
-- AceComm
------------------------------------------------

local COMM_COMPAT = 1
Quixote.commHandlers = {}

function Quixote:OnCommReceive(prefix, sender, distribution, compat, handler, b,c,d,e,f,g,h)
	if compat < COMM_COMPAT then
		self:SendCommMessage("WHISPER", sender, COMM_COMPAT, "OLDVERSION", MAJOR_VERSION, MINOR_VERSION)
	elseif compat > COMM_COMPAT then
		print(sting.format("%s tried to share quest info, but has a newer version of the addon; please upgrade", sender))
	elseif type(handler) == 'string' and self.commHandlers[handler] then
		self.commHandlers[handler](self, sender, b,c,d,e,f,g,h)
	end
end

function Quixote.commHandlers:GETQUESTS(sender)
	self:SyncQuests(sender)
end

function Quixote.commHandlers:QUESTSYNC(sender, quests)
	if self.party[sender] then tableCleaner(self.party[sender])
	else self.party[sender] = {} end
	for title, leaderboard in pairs(quests) do
		self.party[sender][title] = leaderboard
	end
	self:TriggerEvent("Quixote_Party_Quest_Sync", sender)
end

function Quixote.commHandlers:QUESTGAINED(sender, title, leaderboard)
	if not self.party[sender] then self.party[sender] = {} end
	self.party[sender][title] = leaderboard
	self:TriggerEvent("Quixote_Party_Quest_Gained", sender, title)
end

function Quixote.commHandlers:QUESTLOST(sender, title)
	if not self.party[sender] then self.party[sender] = {} end
	self.party[sender][title] = nil
	self:TriggerEvent("Quixote_Party_Quest_Lost", sender, title)
end

function Quixote.commHandlers:LEADERBOARDUPDATE(sender, title, description, numHad, numGot, numNeeded)
	if not self.party[sender] or not self.party[sender][title] then
		self:SendCommMessage("WHISPER", sender, COMM_COMPAT, "GETQUESTS")
	else
		for o,n in pairs(self.party[sender][title]) do
			if o == description then
				self.party[sender][title][o] = numGot
				break
			end
		end
		self:TriggerEvent("Quixote_Party_Leaderboard_Update", sender, title, description, numHad, numGot, numNeeded)
	end
end

function Quixote.commHandlers:OLDVERSION(sender, major, minor)
	print(string.format("You have an older version of the Quixote library than %s -- please upgrade the quest addon which is using it.", sender))
end

function Quixote:Quixote_Leaderboard_Update(name, qid, lbid, description, numHad, numGot, numNeeded, lbtype)
	self:SendCommMessage("PARTY", COMM_COMPAT, "LEADERBOARDUPDATE", name, description, numHad, numGot, numNeeded)
end
function Quixote:Quixote_Quest_Gained(title, id, nObj)
	local l = {}
	for _,g in pairs(self.quests[id].leaderboard) do
		l[g.description] = g.numGot
	end
	self:SendCommMessage("PARTY", COMM_COMPAT, "QUESTGAINED", title, l)
	tableCleaner(l)
end
function Quixote:Quixote_Quest_Lost(title, id)
	self:SendCommMessage("PARTY", COMM_COMPAT, "QUESTLOST", title)
end
function Quixote:PARTY_MEMBERS_CHANGED()
	local p = {}
	for i=1,GetNumPartyMembers(),1 do
		p[UnitName("party"..i)] = true
	end
	for name,_ in pairs(self.party) do
		if not p[name] then
			self.party[name] = nil
		end
	end
	self:SendCommMessage("PARTY", COMM_COMPAT, "GETQUESTS")
	tableCleaner(p)
end
function Quixote:SyncQuests(sender)
	local q = {}
	for quest,id in pairs(self.quest_names) do
		local l = {}
		for _,g in pairs(self.quests[id].leaderboard) do
			l[g.description] = g.numGot
		end
		q[quest] = l
	end
	--Why did this commented out code make "sender"'s game crash?  Why?
	--if sender then
	--	self:SendCommMessage("WHISPER", sender, COMM_COMPAT, "QUESTSYNC", q)
	--else
		self:SendCommMessage("PARTY", COMM_COMPAT, "QUESTSYNC", q)
	--end
	tableCleaner(q)
end

------------------------------------------------
-- API
------------------------------------------------

function Quixote:IsQuestMob(mobname)
	if self.mobs and self.mobs[mobname] then
		local _, _, mob, i, n = string.find(self.mobs[mobname], "(.+): (%d+)/(%d+)")
		return mob, i, n
	end
end

function Quixote:IsQuestItem(itemname)
	if self.items[itemname] then
		local _, _, item, i, n = string.find(self.items[itemname], "(.+): (%d+)/(%d+)")
		return item, i, n
	end
end

function Quixote:IterZones()
	return questIter, self.zones, 0
end

function Quixote:QuestsByZone(zone)
	return questIter, self.quests_zone[zone], 0
end

function Quixote:QuestsByLevel()
	return questIter, self.quests_level, 0
end

function Quixote:GetQuest(q)
	 if type(q)=='number' then return self:GetQuestById(q)
	 else return self:GetQuestByName(q) end
end

function Quixote:GetQuestById(q)
	local q = self.quests[q]
	if q then
		return q.title, q.level, q.tag, q.objective, q.complete, table.getn(q.leaderboard), q.zone, q.id
	end
end

function Quixote:GetQuestByName(q)
	if string.find(q, '^%[') then _,_,q = string.find(q, "^%[?%d*.?]?%s?(.*)") end
	return self:GetQuestById(self.quest_names[q])
end

function Quixote:GetNumQuests()
	return self.numQuests, self.numQuestsDone
end

function Quixote:GetQuestStatusById(q, i)
	if self.quests[q] and self.quests[q].leaderboard[i] then 
		local l = self.quests[q].leaderboard[i]
		return l.description, l.type, l.numGot, l.numNeeded, l.done
	end
end

function Quixote:GetPartyQuestStatus(name, quest, objective)
	if string.find(name, '^party') then name = UnitName(name) end
	if type(quest) == 'number' then quest = self:GetQuestById(quest) end
	if self.party[name] and self.party[name][quest] and self.party[name][quest][objective] then
		return self.party[name][quest][objective]
	end
end

function Quixote:GetReactionLevel(leveltext)
	-- Returns a numeric-equivalent for faction standings.
	if leveltext == FACTION_STANDING_LABEL1 then --Hated
		return 1
	elseif leveltext == FACTION_STANDING_LABEL2 then --Hostile
		return 2
	elseif leveltext == FACTION_STANDING_LABEL3 then --Unfriendly
		return 3
	elseif leveltext == FACTION_STANDING_LABEL4 then --Neutral
		return 4
	elseif leveltext == FACTION_STANDING_LABEL5 then --Friendly
		return 5
	elseif leveltext == FACTION_STANDING_LABEL6 then --Honored
		return 6
	elseif leveltext == FACTION_STANDING_LABEL7 then --Revered
		return 7
	elseif leveltext == FACTION_STANDING_LABEL8 then --Exalted
		return 8
	end
end

------------------------------------------------
-- Register!
------------------------------------------------

AceLibrary:Register(Quixote, MAJOR_VERSION, MINOR_VERSION, activate, nil, external)
